{- $Id$ -}
module Hermite where

{- 高速クォータービュー計算エンジン「Hermite」 -}
{- TODO: 将来は位置計算だけでなく、描画部分も取り込む -}

{-
hermitizeは、与えられた3D地図とカメラ情報から、
カメラ表示範囲内に存在するオブジェクトとその遠さとクォータービュー座標を、
カメラから遠い順にリスト化する。
つまり、hermitizeの結果順に描画していく事で
簡単にクォータービューを表示する事が可能になる。
また、その際に、Hermite管理外のオブジェクトを間に挟む事も可能。
(hermitizeOneで同じフォーマットの結果が得られる。)
尚、三すくみが存在する場合は、その部分にのみ矛盾を持つ結果を返す。
三すくみを正しく解決したい場合は、hermitize2を実装し、使う必要がある。

hermitizeする前に、予め3D地図(または2D地図)を
オブジェクトリストに変換しておく必要がある。
これにはappendMapToObjListを使う。
(3D地図の場合はz座標の数だけこの処理を行う事になる)
渡せる2D地図は、list(y座標)のlist(x座標)に(Maybe [Int])が入ったものとする。
但し、その値がNothingであった場合は、何も存在していなかったものとして扱う。
オブジェクトの数が少ないなら、appendObjToObjListを使うのが効率が良い。
-}

hermitize :: [ ObjInfo ] -> QvCamera -> [ QvObjInfo ]
hermitizeOne :: ObjInfo -> QvCamera -> (Maybe QvObjInfo)

data ObjInfo = ObjInfo ObjId Loc
data ObjId = ObjId (Maybe [ Int ]) -- [ Int ] はユーザが好きに利用可能
data Loc = Loc Float Float Float Float -- (x, y, z, zPrior)
  -- zPriorは、同じ座標に複数のobjが重なった際の優先順位。
  -- zPriorが低いものほど下に表示される。
  -- アイテム、エフェクト、影などの表示に利用できる。

data QvCamera = QvCamera Loc Loc Int Int
  -- (カメラ設置座標, カメラ注視座標, カメラ横幅, カメラ縦幅)
  -- カメラと言いつつ、実際には、該当座標に存在するディスプレイ平面となる。
  -- 横幅と縦幅の単位は半ブロック単位となる。
  -- カメラよりも背後にあるオブジェクトは、距離がマイナスになる。

data QvObjInfo = QvObjInfo ObjId QvLoc
data QvLoc = QvLoc Float Float Float
  -- カメラ視点中央からの、x(横), y(縦), z(奥行き)。
  -- 半ブロック単位。




hermitize objList camera =
  [] -- あとで

hermitizeOne objInfo camera =
  Nothing -- あとで

appendMapToObjList :: [ ObjInfo ] -> Loc -> [[ObjId]] -> [ ObjInfo ]
  -- Locは次に渡す2D地図の原点座標になる。
appendMapToObjList objList originLoc map2d =
  [] -- あとで

appendObjToObjList :: [ ObjInfo ] -> ObjInfo -> [ ObjInfo ]
appendObjToObjList objList newObj = newObj : objList


-- TODO: Map/Obj/Loc/Camera作成補助関数が必要





