module Main (main) where

import Graphics.UI.SDL as SDL
import Graphics.UI.SDL.Rotozoomer
import Graphics.UI.SDL.Primitives
import Graphics.UI.SDL.Framerate as Framerate
import Graphics.UI.SDL.Image as SDLImage
import Graphics.UI.SDL.Mixer as SDLMixer
import Graphics.UI.SDL.TTF as TTF
import System.Exit
import Text.Printf
import Data.Word
import Codec.Binary.UTF8.String



import SDLMisc
import Hermite
import Denzi


caption = "レッツ　ランニング"


main :: IO ()
main = withInit [InitTimer, InitAudio, InitVideo, InitJoystick] $ 
    do screen <- setVideoMode 320 240 0 [SWSurface]
       setCaption (encodeString caption) ""
       withIcon "data/images/icon.bmp" $
         do enableUnicode True
            TTF.init
            openAudio 44100 AudioS16Sys 2 4096
            music <- loadMUS "data/sounds/559.ogg"
            playMusic music (-1)
            bgColor <- mapRGB (surfaceGetPixelFormat screen) 0x00 0x00 0x00
            main2 screen bgColor
            freeMusic music
            closeAudio
            TTF.quit
            return ()

makeHit :: Surface -> IO Surface
makeHit rw =
  do let format = surfaceGetPixelFormat rw
     depth <- pixelFormatGetBitsPerPixel format
     hit <- createRGBSurfaceEndian [SWSurface, SrcAlpha]
                                   24
                                   24
                                   (fromIntegral depth)
     let formatHit = surfaceGetPixelFormat hit
     color <- mapRGBA formatHit 0xff 0x80 0x80 0x80
     fillRect hit Nothing color
     return hit

data MapData = MapData [[(Maybe [Int])]]

main2 :: Surface -> Pixel -> IO ()
main2 rw bgColor =
  do fps <- Framerate.new
     Framerate.set fps 60
     font <- TTF.openFont "data/fonts/day_tt/DAYROM__.ttf" 16
     let floor = Just [0]
         mapData = MapData [[floor, floor, floor, Nothing]
                           ,[floor, floor, Nothing, floor]
                           ,[floor, Nothing, floor, floor]
                           ,[Nothing, floor, floor, floor]
                           ]
     ms <- loadSurfaceWithColorKey "data/images/denzi/Denzi071009-1.png" 255 
     c1 <- loadSurfaceWithColorKey "data/images/denzi/Denzi071022-2.png" 255 
     c2 <- loadSurfaceWithColorKey "data/images/denzi/Denzi071027-6.png" 255 
     se <- loadWAV "data/sounds/kin.wav"
     hit <- makeHit rw
     let transact = transactAllWithCount [(drawClear rw bgColor),
                                          (drawMap rw ms),
                                          (drawChar2 rw c2 se hit),
                                          (drawChar1 rw c1 se hit),
                                          (drawCount rw font)
                                          ]
     loop rw fps transact 0
     freeSurface hit
     freeSurface ms
     freeSurface c1
     freeSurface c2
     TTF.closeFont font
     Framerate.set fps 0
     return ()

loadSurfaceWithColorKey :: String -> Word32 -> IO Surface
loadSurfaceWithColorKey path colorKey =
  do surface <- SDLImage.load path
     lockSurface surface -- TODO: 本当なら、返り値のチェックが必要
     setColorKey surface [SrcColorKey] $ Pixel colorKey
     unlockSurface surface
     return surface



-- 現在の仕様:
-- - マップは一次元。床種別は一種類。
-- - イベントの判定は全てcount依存とする(仮)。
-- - 戦士が右上に歩いていく。
-- - スケルトンが立っている。
-- - 戦う。
-- - 負ける。
drawMap :: Surface -> Surface -> Int -> IO ()
drawMap rw mapSurface count =
  do let floorChipRect = Just (Rect 32 0 32 16)
         eTypeAndCount = getETypeAndCount count -- count値からイベント種別を取得
         floorOffset = getFloorOffset eTypeAndCount -- count値から床オフセットを取得
         pf blockX blockY =
           putFloor rw
                    mapSurface
                    floorChipRect
                    ((160-16) - (16 * blockX) - floorOffset)
                    ((120-8) + (8 * blockY) + (floorOffset `div` 2))
         pfloop x offsets =
           if (null offsets)
              then return ()
              else do pf ((head offsets) - x) ((head offsets) + x)
                      pfloop x (tail offsets)
         len =  [-12..11]
     pfloop (-1) len
     pfloop 0    len
     pfloop 1    len



calcY :: Int -> Int
calcY index = case index of
                   0 -> 0
                   1 -> 80
                   2 -> 160
                   3 -> 224
                   4 -> 306
                   5 -> 368
                   6 -> 450
                   7 -> 512
                   8 -> 594
                   9 -> 240 -- 敵の攻撃用
                   _ -> 0





getChar2Location :: (EType, Int) -> Int -> (Int, Int)
getChar2Location pair count =
  case (fst pair) of
    ESTOP -> (0, 0) 
    EMOVE -> (0, 0) 
    EMOVEWITHENEMY -> (320 - count, (-24) + (count `div` 2))
    EFIGHT1 -> (160, 120-64)
    EFIGHT2 -> (160, 120-64)
    EDEAD -> (160, 120-64)


drawChar2 :: Surface -> Surface -> Chunk -> Surface -> Int -> IO ()
drawChar2 rw charSurface se hit count =
  do let eTypeAndCount = getETypeAndCount count
         charPattern = getChar2Pattern eTypeAndCount
         charX = (snd charPattern) * 64
         charY = calcY (fst charPattern)
         countTrue = (snd eTypeAndCount)
         (locX, locY) = getChar2Location eTypeAndCount countTrue
     (aX, aY) <- case count of
                   1208 -> do playChannel (-1) se 0
                              return ((-2), 1)
                   1209 -> return ((-4), 2)
                   1210 -> return ((-2), 1)
                   1272 -> do playChannel (-1) se 0
                              return ((-2), 1)
                   1273 -> return ((-4), 2)
                   1274 -> return ((-2), 1)
                   1336 -> do playChannel (-1) se 0
                              return ((-2), 1)
                   1337 -> return ((-4), 2)
                   1338 -> return ((-2), 1)
                   1400 -> do playChannel (-1) se 0
                              return ((-2), 1)
                   1401 -> return ((-4), 2)
                   1402 -> return ((-2), 1)
                   _ -> return (0, 0)
     if (charY == 0)
        then return True
        else blitSurface charSurface
                         (Just (Rect charX charY 64 64))
                         rw
                         (Just (Rect (locX-aX) (locY-aY) 0 0))
     if (aX == 0)
        then return True
        else blitSurface hit Nothing rw (Just (Rect (locX+20-aX)
                                              (locY+20-aY)
                                              0
                                              0))
     return ()


drawChar1 :: Surface -> Surface -> Chunk -> Surface -> Int -> IO ()
drawChar1 rw charSurface se hit count =
  do let eTypeAndCount = getETypeAndCount count
         charPattern = getChar1Pattern eTypeAndCount
         charX = (snd charPattern) * 64
         charY = calcY (fst charPattern)
     (aX, aY) <- case count of
                   1236 -> do playChannel (-1) se 0
                              return ((-2), 1)
                   1237 -> return ((-4), 2)
                   1238 -> return ((-2), 1)
                   1300 -> do playChannel (-1) se 0
                              return ((-2), 1)
                   1301 -> return ((-4), 2)
                   1302 -> return ((-2), 1)
                   1364 -> do playChannel (-1) se 0
                              return ((-2), 1)
                   1365 -> return ((-4), 2)
                   1366 -> return ((-2), 1)
                   1428 -> do playChannel (-1) se 0
                              return ((-2), 1)
                   1429 -> return ((-4), 2)
                   1430 -> return ((-2), 1)
                   1440 -> do SDLMixer.fadeOutMusic 2000
                              return (0, 0)
                   _ -> return (0, 0)
     blitSurface charSurface
                 (Just (Rect charX charY 64 64))
                 rw
                 (Just (Rect (160-32+aX) (120-48+aY) 0 0))
     if (aX == 0)
        then return True
        else blitSurface hit Nothing rw (Just (Rect (160-32+20+aX)
                                              (120-48+20+aY)
                                              0
                                              0))
     return ()


data EType = ESTOP -- 歩いてるだけ
           | EMOVE -- 立ち止まる
           | EMOVEWITHENEMY -- 向こうに敵が立っている
           | EFIGHT1 -- 戦闘中/こっちの攻撃
           | EFIGHT2 -- 戦闘中/敵の攻撃
           | EDEAD -- 死亡


-- XXX: 諸事情により、イベント判定カウントは32の倍数である事とする
getETypeAndCount :: Int -> (EType, Int)
getETypeAndCount count | (count < 128) = (ESTOP, 0)
                       | (count < 1024) = (EMOVE, count - 128)
                       | (count < 1184) = (EMOVEWITHENEMY, count - 1024)
                       | (count < 1216) = (EFIGHT1, count - 1184)
                       | (count < 1248) = (EFIGHT2, count - 1216)
                       | (count < 1280) = (EFIGHT1, count - 1248)
                       | (count < 1312) = (EFIGHT2, count - 1280)
                       | (count < 1344) = (EFIGHT1, count - 1312)
                       | (count < 1376) = (EFIGHT2, count - 1344)
                       | (count < 1408) = (EFIGHT1, count - 1376)
                       | (count < 1440) = (EFIGHT2, count - 1408)
                       | otherwise = (EDEAD, count - 1440)


getFloorOffset :: (EType, Int) -> Int
getFloorOffset pair =
  let moveCount = ((snd pair) `mod` 16)
  in case (fst pair) of
       ESTOP -> 0
       EMOVE -> moveCount
       EMOVEWITHENEMY -> moveCount
       EFIGHT1 -> 0
       EFIGHT2 -> 0
       EDEAD -> 0


-- 特例として、1430以降は(8, 0)を返す
getChar1Pattern :: (EType, Int) -> (Int, Int)
getChar1Pattern pair =
  case (fst pair) of
    ESTOP -> (1, 2)
    EMOVE -> (3, ((snd pair) `mod` 32) `div` 8)
    EMOVEWITHENEMY -> (3, ((snd pair) `mod` 32) `div` 8)
    EFIGHT1 -> (5, ((snd pair) `mod` 32) `div` 8)
    EFIGHT2 -> if (1430 <= (snd pair))
                  then (8, 0)
                  else (1, 2)
    EDEAD -> (8, 0)

getChar2Pattern :: (EType, Int) -> (Int, Int)
getChar2Pattern pair =
  case (fst pair) of
    ESTOP -> (0, 0)
    EMOVE -> (0, 0)
    EMOVEWITHENEMY -> (1, 1)
    EFIGHT1 -> (1, 1)
    EFIGHT2 -> (9, (((snd pair) `mod` 32) `div` 8) + 4)
    EDEAD -> (1, 1)



putFloor :: Surface -> Surface -> (Maybe Rect) -> Int -> Int -> IO ()
putFloor rw mapSurface floorChipRect x y =
  do blitSurface mapSurface
                 floorChipRect
                 rw
                 (Just (Rect x y 0 0))
     return ()



drawClear :: Surface -> Pixel -> Int -> IO ()
drawClear rw bgColor count =
  do fillRect rw Nothing bgColor
     return ()


transactAllWithCount :: [(Int -> IO ())] -> Int -> IO ()
transactAllWithCount procs count =
  if (null procs)
     then return ()
     else do (head procs) count
             transactAllWithCount (tail procs) count

drawCount :: Surface -> Font -> Int -> IO ()
drawCount target font count =
  do let format = surfaceGetPixelFormat target
         bgColor = (Color 0x00 0x00 0x00)
         fgColor = (Color 0xFF 0xFF 0xFF)
         str = printf "Count: %d" count
     depth <- pixelFormatGetBitsPerPixel format
     text <- renderUTF8Shaded font (encodeString str) fgColor bgColor
     blitSurface text Nothing target Nothing
     freeSurface text
     return ()


withIcon :: String -> IO () -> IO ()
withIcon path thunk
  = do icon <- loadSurfaceWithColorKey path 15
       setIcon icon Nothing
       freeSurface icon -- プロセス終了時まで解放しない方がいいのかどうかは謎
       thunk



loop :: Surface -> FPSManager -> (Int -> IO ()) -> Int -> IO ()
loop rw fps proc count
  = do event <- pollEvent
       case event of
         Quit -> exitWith ExitSuccess
         KeyDown (Keysym SDLK_F4 [KeyModLeftAlt] _) -> exitWith ExitSuccess
         KeyDown (Keysym SDLK_F4 [KeyModRightAlt] _) -> exitWith ExitSuccess
         _ -> proc count
       Framerate.delay fps
       SDL.flip rw
       loop rw fps proc (count + 1)




